/* *****************************************************************************
 * Copyright (c) IRISA Laboratory, 
 * Universite Bretagne Sud, France
 * <http://www-casa.irisa.fr/>
 * ****************************************************************************/

/* 
 * This module provides means to implement a listener to DoDWAN events 
 * related to changes its neighborhood.
 * 
 * A user can 
 *  - ask DoDWAN for the current peers (getPeers())
 *  - register a listener that will be notified from DoDWAN for the add, remove and clear event from DoDWAN
 * 
 * As some events may be missed (to be invesigated), additionnaly, the service maintains its own set of known peers
 * and  
 *
 */


console.log("Start module peer.js");

import * as napiConn from "./connection.js"
import * as util from "./util.js"
import * as napiSession from "./session.js"


const PERIOD = 12000; // Get peers every 10s



function log(msg) {
    return;
    util.serviceLog(null, 'napi_peer', msg);
}


/*
 * Class PeerService ---------------------------------------------
 * The Peer service is attached to an existing session.
 */
export class PeerService {

    constructor(attachedSession) {
                
        this.peerSet; //Set of known peers
        this.timeoutId;  // Handler for the delayed task of periodic polling 
        this.intervalId; // Handler of the periodic task that update the peer set
        this.session = attachedSession;
        this.listeners = new Set(); // set of peerListeners to notify
        
        this.sessionListener = new napiSession.SessionListener(null, null,  null, null, () => {this.start();}, () => {this.stop();});
        this.session.addListener(this.sessionListener);
        this.session.setOnReceivePeerNotification((pdu) => {this.receiveNotif(pdu);});   
    }
    
    /**
     * Starts working (after the session is started)
     * Programs the periodic polling of peers. 
     * It starts (after PERIOD ms) updating the set of known peer every PERIOD ms.
     * @returns {undefined}
     */
    start() {
        console.log("Start polling peers in "+PERIOD/1000+ " sec");
                        
        this.timeoutId = setTimeout ( 
                () => {
                       this.intervalId = setInterval( () => {this.updateKnownPeers();},  PERIOD );
                      }, 
                PERIOD, 
        );
                       
    }
    
    
    
    /**
     * Asks Dodwan for the list of current peers, and update the set of known peers.
     * Trigger the potential addPeer() and removePeer() events according to what has been learned
     * @returns {undefined}
     */
    updateKnownPeers() {
        //console.log("Update known peers");
        this.getPeers(
                (peersArray) => {
                   const newPeerSet = new Set(peersArray);
           
                   for (const peer of this.peerSet) { 
                        if (! newPeerSet.has(peer)) {
                           for (const listener of this.listeners) {
                               listener.removePeer(peer);
                           }
                        }      
                   }
                    
                   for (const peer of newPeerSet) { 
                        if (! this.peerSet.has(peer)) {
                           for (const listener of this.listeners) {
                               listener.addPeer(peer);
                           }
                        }      
                   } 
                             
                });
    }
    
    
    
    /**
     * stops working (when the session is stopped)
     * @returns {undefined}
     */
    stop() {
        console.log("Stop polling peers");
        clearInterval(this.intervalId);
        clearTimeout(this.timeoutId);
        this.session.removeListener(this.sessionListener);
    }
    
    
    
    
    
    
    /**
    * Gives the id of the local peer (host name)
    * @param {function(string)} the callback passing the local peer id
    */
    getMyPeerName(okcallback, errorcallback) { 
        let command = new napiConn.Pdu();
        command.name = "get_my_peer";
        command.tkn = util.uniqueId();
        //log("PeerService sending get_my_peer" + JSON.stringify(command));
        this.session.sendCommand(command, 
            (pdu) => {
                if (pdu.name !== "recv_my_pid") {
                  console.error("Strange: response to get_my_peer is not recv_my_pid");
                  return;
                }
                okcallback(pdu.pid);
            },
          
            (error) => { 
                console.error("PeerService: error when issuing get_my_peer"); 
                if (errorcallback) errorcallback(error);
            }
        );
    }
    
    
    
    /**
    * Gives the ids of the peers that are currently in the direct neighborhood
    * @param {function(string[])} the callback passing the array of ids
    */
    getPeers(okcallback, errorcallback) {
        let command = new napiConn.Pdu();
        command.name = "get_peers";
        command.tkn = util.uniqueId();
        //log("PeerService sending get_peers: " + JSON.stringify(command));
        this.session.sendCommand(command, 
            (pdu) => {
                if (pdu.name !== "recv_pids") {
                  console.error("Strange: response to get_peers is not recv_pids");
                  return;
                }
                this.peerSet = new Set(pdu.pids);
                okcallback(pdu.pids);
            },
          
            (error) => { 
                console.error("PeerService: error when issuing get_peers"); 
                if (errorcallback) errorcallback(error);
            }
        );
    
    }
    
    
    /**
    * Register a listener to observe the peers in the direct neighborhood
    *
    * @param {PeerListener} listener - the listener to be registered
    */
    addListener(listener) {  
       this.listeners.add(listener);
    }


    /**
    * Unregister a listener
    *
    * @param {PeerListener} listener - the listener to be unregistered
    */
    removeListener(listener) { 
        this.listeners.delete(listener);
    }
    
    
    
     /**
     * Handles the reception of a peer notification (PDU without token).
     * @param {Pdu} notifPdu - the received PDU 
     */
    receiveNotif(notifPdu) {
        // ..... add_peer ..........................................................
        if (notifPdu.name === "add_peer") {
            log("Peer service received add_peer |pid=" + JSON.stringify(notifPdu.pid) + "|");
            if (! this.peerSet.has(notifPdu.pid)) {
                this.peerSet.add(notifPdu.pid);
                for (const listener of this.listeners) {
                    listener.addPeer(notifPdu.pid);
                }
            }
            return;
        }
        // ..... remove_peer ..........................................................
        else if (notifPdu.name === "remove_peer") {
            log("Peer service received remove_peer |pid=" + JSON.stringify(notifPdu.pid) + "|");
            if (this.peerSet.has(notifPdu.pid)) {
                this.peerSet.delete(notifPdu.pid);
                for (const listener of this.listeners) {
                    listener.removePeer(notifPdu.pid);
                }
            }
            return;
        }
        // ..... clear_peers ..........................................................
        else if (notifPdu.name === "clear_peers") {
            log("Peer service received clear_peers");
            this.peerSet.clear();
            for (const listener of this.listeners) {
                listener.clearPeers();
            }
            return;
        }
    }
} 
    

    

/*
 * Interface PeerListener ---------------------------------------------
 */
export class PeerListener {

    constructor(addPeer, removePeer, clearPeers) {
        this.addPeer = addPeer;
        this.removePeer = removePeer;
        this.clearPeers = clearPeers;
    }
    

}

