/* *****************************************************************************
 * Copyright (c) IRISA Laboratory, 
 * Universite Bretagne Sud, France
 * <http://www-casa.irisa.fr/>
 * ****************************************************************************/

/* 
 * This module provides 
 *   - a Session object, named service, that offers methods and callabcks for
 *        . handling the lifecycle of a session 
 *        . sending commands, and receiving responses and notifications
 *   - Interface SessionListener to receive events related to the status of the session
 */

console.log("Start module session.js");

import * as util from "./util.js"
import * as napiConn from "./connection.js"
import * as napiConf from "./configuration.js"


function log(msg) {
    return;
    util.serviceLog(null, 'napi_session', msg);
}

 
/*
 * Class Session ---------------------------------------------
 */
export class Session {
    

    constructor(conf) {
 
        this.configuration = conf;
 
        // List of subscriptions registered in DoDWAN
        this.subscriptions = [];
        
        this.client = util.persistentUniqueId(); 
        
        this.continuation = false;
        
        this.reconnectPeriod = 15;
        this.reconnectNbTriesMax = 5;
        
        this.listeners = new Set();
        
      
        this.onReceivePubsubNotification = function () {log("Empty onReceivePubsubNotification");};
        this.onReceivePeerNotification = function () {log("Empty onReceivePeerNotification");};
        this.onReceiveD2dNotification = function () {log("Empty onReceiveD2dNotification");};
      
      
        // Maps of the callbacks to execute when receiving a response or an error,
        // for each token.  A callback should be removed just after its execution
        // TODO : garbage collect callbacks that may linger after a while 
        this.responseCallbacks = new Map();
        this.errorCallbacks = new Map();
        
        // Handles incoming PDUS (notifications and responses)
        this.connection = new napiConn.WebsocketConnection();
        this.connection.setOnReceivePdu((pdu) => {
            this.receivePdu(pdu);
        });
      
    }
    
    
   /*-----------------------------------------------------------
    * Sets a callback for pubsub notifications
    * @param {function} f - the callback to be registered
    */
    setOnReceivePubsubNotification(f) { 
        this.onReceivePubsubNotification = f; 
    }
     
    /*-----------------------------------------------------------
    * Sets a callback for peer notifications (appeareance and disapperance of peers)
    * @param {function} f - the callback to be registered
    */    
    setOnReceivePeerNotification(f) { 
        this.onReceivePeerNotification = f; 
    }
     
    /*-----------------------------------------------------------
    * Sets a callback for 2D2 notifications (messages issued by neighbor peers)
    * @param {function} f - the callback to be registered
    */    
    setOnReceiveD2dNotification(f) { 
        this.onReceiveD2dNotification = f; 
    }
    
   /*-----------------------------------------------------------
    * Registers a session listener, to observe the changes in the status of the session
    * @param {SessionListener} listener - the listener to be registered
    */
    addListener(listener) {
       this.listeners.add(listener);
    }


   /*-----------------------------------------------------------
    * Unregisters a session listener
    * @param {SessionListener} listener - the listener to be unregistered
    */
    removeListener(listener) { 
        this.listeners.delete(listener);
    }
    
    
    /*-----------------------------------------------------------
     * Tries to reconnect periodically to the server
     * @param {string} uri - the server uri (ws://host:port/path/) 
     * @private
     */
    reconnectLater(uri) {
        this.continuation = true;
        if ((this.reconnectNbTriesMax === 0) ||  (this.reconnectPeriod === 0)) return;
        this.reconnectNbTries ++;     
          if (this.reconnectNbTries < this.reconnectNbTriesMax) {
             log("Will try to reconnect in "+this.reconnectPeriod+"s (try #" + this.reconnectNbTries + ")");
             util.setTimeout(() => {this.connection.connect(uri);}, this.reconnectPeriod*1000);
          }
          else {
              log("Give up trying to reconnect"); 
              for (const listener of this.listeners) listener.onSessionStopped();
          }
    }


    
    /*-----------------------------------------------------------
    * Closes the session. All subscriptions are forgotten and the connection is closed
    */
    bye() {
        this.subscriptions = [];
        this.connection.disconnect();  
        for (const listener of this.listeners) listener.onSessionStopped();
    }
    
    
    
     /*-----------------------------------------------------------
     * Asks to be ready to publish and receive notifications: 
     * open a connection, and when it is opened, tells the server to open a session, and start notifications
     */
    start() {
       
        let uri = "ws://" + this.configuration.config.server_host + ":" + this.configuration.config.server_port
                   + "/" + this.configuration.config.name;
        this.reconnectPeriod = this.configuration.config.reconnect_period;
        this.reconnectNbTriesMax = this.configuration.config.reconnect_nb_tries_max;
        
        // Begin in the starting state
        for (const listener of this.listeners) listener.onSessionStarting();      
        
        this.connection.setOnSocketError((error) => {
            for (const listener of this.listeners) listener.onConnectionError(error);
            this.reconnectLater(uri);   
        });
        
        this.connection.setOnSocketClose(() => {
            for (const listener of this.listeners) listener.onConnectionClosed();
            this.reconnectLater(uri);       
        });
        
        
        this.connection.setOnSocketOpen(() => {
            
            // Opens a session 
            let command = new napiConn.Pdu();
            command.name = "hello";
            command.client = this.client;
            command.cont = this.continuation;
            command.tkn = util.uniqueId();
            this.sendCommand(command,
                () => {
                    // OK : authorize traffic to and from DoDWAN
                    let command = new napiConn.Pdu();
                    command.name = "start";
                    command.tkn = util.uniqueId();
                    this.sendCommand(command,
                         (okPdu) => { log("Traffic to/from DoDWAN authorized");
                                      for (const listener of this.listeners) listener.onSessionStarted();
                                    },
                             
                         (error) => {console.error("Strange error when authorizing traffic :"+error);
                                     bye(); // error considered fatal
                                    }
                    );
                  
                  
                  
                  
                },
                
                (error) => {console.error("Strange error when opening session : "+error);
                            bye(); // error considered fatal
                           }
            );
            
            // Trigger the listeners
            for (let listener of this.listeners) listener.onConnectionOpened();
            
        });
        
        
        // Initial connection
        this.connection.connect(uri); 
         
    }
    
    
    
    /*-----------------------------------------------------------
    * Stops the traffic from/to DoDWAN
    */
    stop() {
        let command = new napiConn.Pdu();
        command.name = "stop";
        command.tkn = util.uniqueId();
        sendCommand(command,
            (response) => { log("Traffic to/from DoDWAN blocked");
                    for (const listener of this.listeners) listener.onSessionStopped();
                  },
                             
                  (error) => {console.error("Strange error when blocking traffic :"+error);
                              bye(); // error considered fatal
                  }
        );            
    }
    
    
    
    /*-----------------------------------------------------------
     * Sends a command PDU
     * @param {Pdu} pdu - The PDU to be sent
     * @param {function(Pdu)} responseCallback - the function passing the response to call when all goes well 
     * @param {function(string)} errorCallback - the function passing the error reason when an error occurs 
     */ 
    sendCommand(pdu, responseCallback, errorCallback) {
       if (pdu.tkn) {
           if (responseCallback) this.responseCallbacks.set(pdu.tkn, responseCallback); 
           if (errorCallback) this.errorCallbacks.set(pdu.tkn, errorCallback);
        }
        this.connection.sendPdu(pdu);
        return;
    }
    


    //////////////////////////////////////////////////////////////////////////////////////
    // Private methods
    //////////////////////////////////////////////////////////////////////////////////////

    /*-----------------------------------------------------------
     * Dispatch the received PDU to the good notifier
     * @param {Pdu} pdu - The received PDU
     * @private
     */ 
    receivePdu(pdu) {
        if (pdu.tkn) {
            this.receiveResponse(pdu);
        }
        else {
            this.receiveNotification(pdu);
        }
    }
    
      

    /*-----------------------------------------------------------
     * Handles a response PDU: triggers the ok or error callback
     * corresponding to the command, according to the token present in the PDU
     * @param {Pdu} pdu - The received response PDU
     * @private
     */
    receiveResponse(pdu) {
        if (pdu.name === "error") {
            let errorCallback = this.errorCallbacks.get(pdu.tkn);
            if (errorCallback !== undefined) {
                errorCallback(pdu.reason);
                this.errorCallbacks.delete(pdu.tkn);
            }
            
        } else { // any other name, including ok
            let responseCallback = this.responseCallbacks.get(pdu.tkn);
            if (responseCallback !== undefined) {
               responseCallback(pdu);
               this.responseCallbacks.delete(pdu.tkn);
            }   
        }
    }
     
     
   /*-----------------------------------------------------------
    * Dispatch a notification PDU to the right pre-registered notifier:
    * (pubsub or peer notifier)
    * @param {Pdu} pdu - The received notification PDU
    * @private
    */
    receiveNotification(pdu) {
        
        if (pdu.name === "recv_desc") {
            //log("Session received pubsub notification PDU "+JSON.stringify(pdu));
            this.onReceivePubsubNotification(pdu);
        }
        else if ((pdu.name === "add_peer") || (pdu.name === "remove_peer") || (pdu.name === "clear_peers")){
            this.onReceivePeerNotification(pdu);
        }
        else if (pdu.name === "recv_from_peer") {
            this.onReceiveD2dNotification(pdu);
        }
        else {
            console.error("Unknown notification "+pdu.name);
        }
    }
     
     
    
  

}





/*
 * Interface SessionListener ---------------------------------------------
 */
export class SessionListener {

    constructor( onConnectionOpened, onConnectionClosed, onConnectionError, 
                 onSessionStarting, onSessionStarted, onSessionStopped) {
                     
    /**
    * Method called when the connection associated to the session is opened
    */             
    this.onConnectionOpened = (onConnectionOpened === null ) ? ()=>{} : onConnectionOpened;
        
    /**
    * Method called when the connection associated to the session is closed
    */
    this.onConnectionClosed = (onConnectionClosed === null ) ? ()=>{} : onConnectionClosed;

    /**
    * Method called when an error occurs while opening the connection associated
    * to the session
    * @param {string} reason 
    */
    this.onConnectionError = (onConnectionError === null ) ? ()=>{} : onConnectionError;

    /**
    * Method called when the session is being starting: trying to (re)open a connection
    * and transfer opening pdus to the server
    */
    this.onSessionStarting = (onSessionStarting === null ) ? ()=>{} : onSessionStarting;

    /**
    * Method called when the session is started
    */
    this.onSessionStarted = (onSessionStarted === null ) ? ()=>{} : onSessionStarted;

    /**
    * Method called when the session is stopped. Can occur either after a
    * {@link Session#stop()} invocation or when the connection cannot be established
    * or reestablished after having been disconnected 
    */
    this.onSessionStopped = (onSessionStopped === null ) ? ()=>{} : onSessionStopped;
    }
    

}

